package com.guildasoft.bluechatroom2

import android.content.Context
import android.graphics.Color
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.TextView
import androidx.recyclerview.widget.RecyclerView
import com.guildasoft.bluechatroom2.entity.AppMessage
import com.guildasoft.bluechatroom2.entity.Constants
import java.text.SimpleDateFormat
import java.util.*

class ChatAdapter(val chatData: List<AppMessage>, val context: Context) : RecyclerView.Adapter<RecyclerView.ViewHolder>() {

    val SENT = 0
    val RECEIVED = 1
    var df: SimpleDateFormat = SimpleDateFormat("hh:mm a",Locale.getDefault())

    override fun onBindViewHolder(p0: RecyclerView.ViewHolder, position: Int) {

        when(p0.itemViewType){

            SENT -> {
                val holder: SentHolder = p0 as SentHolder
                holder.sentTV.text = chatData[position].text
                val timeMilliSeconds = chatData[position].time
                val calendar = Calendar.getInstance()
                calendar.timeInMillis = timeMilliSeconds.toLong()
                holder.timeStamp.text = df.format(calendar.time)
            }
            RECEIVED -> {
                val holder: ReceivedHolder = p0 as ReceivedHolder
                holder.receivedTV.text = chatData[position].text
                val timeMilliSeconds = chatData[position].time
                val calendar = Calendar.getInstance()
                calendar.timeInMillis = timeMilliSeconds.toLong()
                holder.timeStamp.text = df.format(calendar.time)
                holder.nickname.text = chatData[position].owner
                val color = Color.RED
                holder.nickname.setTextColor(color)
            }

        }
    }

    override fun getItemViewType(position: Int): Int {

        when(chatData[position].type){
            Constants.MESSAGE_TYPE_SENT -> return SENT
            Constants.MESSAGE_TYPE_RECEIVED -> return RECEIVED
        }

        return -1
    }

    override fun getItemCount(): Int {
        return chatData.size
    }

    override fun onCreateViewHolder(p0: ViewGroup, viewType: Int): RecyclerView.ViewHolder {

        when(viewType){
            SENT -> {
                val view = LayoutInflater.from(context).inflate(R.layout.sent_layout,p0,false)
                return SentHolder(view)
            }
            RECEIVED -> {
                val view = LayoutInflater.from(context).inflate(R.layout.received_layout,p0,false)
                return ReceivedHolder(view)
            }
            else -> {
                val view = LayoutInflater.from(context).inflate(R.layout.sent_layout,p0,false)
                return SentHolder(view)
            }
        }
    }

    inner class SentHolder(itemView: View) : RecyclerView.ViewHolder(itemView){
        var sentTV = itemView.findViewById<TextView>(R.id.sentMessage)
        var timeStamp = itemView.findViewById<TextView>(R.id.timeStamp)
    }

    inner class ReceivedHolder(itemView: View) : RecyclerView.ViewHolder(itemView) {
        var receivedTV = itemView.findViewById<TextView>(R.id.receivedMessage)
        var timeStamp = itemView.findViewById<TextView>(R.id.timeStamp)
        var nickname = itemView.findViewById<TextView>(R.id.nickName)
    }

}