package com.guildasoft.bluechatroom2;

import android.Manifest;
import android.annotation.SuppressLint;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.os.Build;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;
import android.view.View;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.appcompat.app.AlertDialog;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.app.ActivityCompat;
import androidx.core.content.ContextCompat;
import androidx.fragment.app.FragmentManager;
import androidx.fragment.app.FragmentTransaction;

import com.guildasoft.bluechatroom2.adapter.BluetoothChatService;
import com.guildasoft.bluechatroom2.entity.AppMessage;
import com.guildasoft.bluechatroom2.entity.Constants;
import com.guildasoft.bluechatroom2.entity.Device;
import com.guildasoft.bluechatroom2.utils.MessageUtils;
import com.guildasoft.bluechatroom2.utils.MessageUtilsImpl;
import com.guildasoft.bluechatroom2.utils.MessageUtilsImpl.MessageTypes;

import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Timer;
import java.util.TimerTask;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

public class MainActivity extends AppCompatActivity implements ChatFragment.CommunicationListener {
    private static final int MESSAGE_UUID_LIFETIME_MS = 15000;
    private static final int FINE_LOCATION_RQ = 100;
    private static final int COARSE_LOCATION_RQ = 101;
    private static final int BLUETOOTH_RQ = 102;
    private static final int BLUETOOTH_ADMIN_RQ = 103;
    private static final int ENABLE_BLUETOOTH_CODE = 104;
    private static final String IDENTIFIER_APP = "*-*";
    private static final int MAX_NAME_SIZE = 20;

    List<String> recentUuids = new ArrayList<>();
    private Timer timer;
    BluetoothAdapter bluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
    List<Device> chatDevices = new ArrayList<>();
    List<String> onlineDevices = new ArrayList<>();
    Handler blueChatHandler;
    BluetoothChatService mainChatService;
    String bluetoothOriginalName;
    Context mainActivityContext;

    private final ChatFragment chatFragment = new ChatFragment();
    private MainButtonsFragment mainButtonsFragment;

    MessageUtils converter = new MessageUtilsImpl();

    @SuppressLint("HandlerLeak")
    private Handler createHandlerCope(){
        return new Handler(){
            @Override
            public void handleMessage(@NonNull Message msg) {
                if(msg == null) return;
                if(msg.what == Constants.Companion.getMESSAGE_STATE_CHANGE()){
                    if (msg.arg1 == BluetoothChatService.Companion.getNEW_INCOMING_CONNECTION()){
                        BluetoothSocket newSocket = (BluetoothSocket) msg.obj;
                        BluetoothDevice newConnection = newSocket.getRemoteDevice();
                        Device newDevice = Device.builder()
                                .bluetoothDevice(newConnection)
                                .nickName(removeIdentifier(newConnection.getName()))
                                .chatService(new BluetoothChatService(mainActivityContext, blueChatHandler))
                                .build();
                        chatDevices.add(newDevice);
                        String socketType = msg.arg2==1?"Secure":"Insecure";
                        newDevice.getChatService().connected(newSocket, newSocket.getRemoteDevice(),socketType);
                        if(onlineDevices.contains(newConnection.getName())){
                            sendPermissionDeniedError(newDevice);
                            newDevice.getChatService().stop();
                            removeNeighborFromLists(newDevice.getNickName());
                        }
                    }
                    if (msg.arg1 == BluetoothChatService.Companion.getSTATE_CONNECTED()){
                        showToastText("CONECTADO \uD83E\uDD73");
                        sendConnectedMessage();
                        showChatFragment();
                    }
                    if (msg.arg1 == BluetoothChatService.Companion.getSTATE_CONNECTING()){
                        showToastText("Quase lá ⌛");
                    }
                    if (msg.arg1 == BluetoothChatService.Companion.getSTATE_LISTEN()
                            || msg.arg1 == BluetoothChatService.Companion.getSTATE_NONE()){
                        showToastText("Sem conexão \uD83D\uDE25");
                    }
                }

                if(msg.what == Constants.Companion.getMESSAGE_WRITE()){
                    String writeMessage = new String((byte[]) msg.obj, StandardCharsets.UTF_8);
                    String milliseconds = String.valueOf(System.currentTimeMillis());
                    if(!recentUuids.contains(converter.getUuidFromMessage(writeMessage))){
                        MessageTypes type = converter.getTypeFromMessage(writeMessage);
                        if(type.equals(MessageTypes.CHAT_ORIGIN) || type.equals(MessageTypes.CHAT_SHARE)) {
                            chatFragment.communicate(
                                    AppMessage.builder()
                                            .text(converter.getTextFromMessage(writeMessage))
                                            .time(milliseconds)
                                            .type(Constants.Companion.getMESSAGE_TYPE_SENT())
                                            .build()
                            );
                        }
                    }
                    saveUuidByMessageAndCreateTaskAutoRemove(writeMessage);
                }

                if(msg.what == Constants.Companion.getMESSAGE_READ()){
                    AppMessage message = (AppMessage) msg.obj;
                    if(!recentUuids.contains(message.getUuid())){
                        if(message.getMessageType().equals(MessageTypes.CHAT_ORIGIN) || message.getMessageType().equals(MessageTypes.CHAT_SHARE)){
                            message.setType(Constants.Companion.getMESSAGE_TYPE_RECEIVED());
                            chatFragment.communicate(message);
                            distributeMessage(converter.mountMessageUsingAppMessage(message));
                        }
                        if(message.getMessageType().equals(MessageTypes.INFO_FAREWELL)){
                            showToastText(message.getText()+" saiu");
                            removeNeighborFromLists(removeIdentifier(message.getText()));
                        }
                        if(message.getMessageType().equals(MessageTypes.INFO_ARRIVAL)){
                            onlineDevices.add(removeIdentifier(message.getOwner()));
                        }
                        if(message.getMessageType().equals(MessageTypes.ERROR)){
                            showToastText(message.getText());
                        }
                    }
                    saveUuidInRecentsAndCreateTaskAutoRemove(message.getUuid());
                }

                if (msg.what == Constants.Companion.getMESSAGE_TOAST()){
                    showToastText(msg.getData().getString(Constants.Companion.getTOAST()));
                }

                if (msg.what == Constants.Companion.getMESSAGE_CONNECTION_LOST()){
                    String name = msg.getData().getString(Constants.Companion.getCONNECTION_LOST());
                    if(Objects.nonNull(name) && !name.isEmpty()) {
                        removeNeighborFromLists(removeIdentifier(name));
                        showToastText(name + " saiu");
                        informDeviceExit(name);
                        if(chatDevices.isEmpty()){
                            showAlertEmptyChat();
                        }
                    }
                }
            }
        };
    }

    private void sendPermissionDeniedError(Device device){
        AppMessage message = AppMessage.builder()
                .messageType(MessageTypes.ERROR)
                .owner(removeIdentifier(bluetoothAdapter.getName()))
                .uuid(generateNewUuid())
                .text("Acesso negado, o nome de usuário já existe!")
                .build();
        device.getChatService().sendMessage(message);
    }

    private void showAlertEmptyChat(){
        new AlertDialog.Builder(this)
                .setTitle("Perda de conexão")
                .setMessage("Parece que você perdeu completamente a conexão com a rede. Tente encontrar outros dispositivos.")
                .setPositiveButton("Buscar", (dialog, id) -> {
                    findDevices(null);
                })
                .setNegativeButton("Fechar", (dialog, id) -> {})
                .show();
    }

    private void informDeviceExit(String nameGone){
        AppMessage messageToSend = AppMessage.builder()
                .messageType(MessageTypes.INFO_FAREWELL)
                .uuid(generateNewUuid())
                .owner(getFirstNameWithoutIdentifier())
                .text(removeIdentifier(nameGone))
                .build();
        chatDevices.forEach(
                device -> device.getChatService().sendMessage(messageToSend)
        );
    }

    private void removeNeighborFromLists(String name){
        onlineDevices.remove(name);
        chatDevices = chatDevices.stream().filter(device -> !device.getNickName().equals(name)).collect(Collectors.toList());
    }

    private void distributeMessage(String message){
        String uuid = converter.getUuidFromMessage(message);
        String text = converter.getTextFromMessage(message);
        String name = converter.getUserFromMessage(message);
        String reassembleMessage = converter.mountCompleteMessage(MessageTypes.CHAT_SHARE, uuid, name, text);
        chatDevices.forEach(
            device -> device.getChatService().write(reassembleMessage.getBytes())
        );
    }

    private void sendConnectedMessage(){
        String uuid = generateNewUuid();
        String mountedMessage = converter.mountCompleteMessage(
                MessageTypes.INFO_ARRIVAL, uuid,
                getFirstNameWithoutIdentifier(),
                getFullNameWithoutIdentifier()
        );
        chatDevices.forEach(
            device -> device.getChatService().write(mountedMessage.getBytes())
        );
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);
        mainActivityContext = this;

        blueChatHandler = createHandlerCope();
        mainChatService = new BluetoothChatService(this, blueChatHandler);
        checkBluetooth();

        bluetoothOriginalName = bluetoothAdapter.getName();
        if(!containsIdetifier(bluetoothOriginalName)) {
            String newName = bluetoothOriginalName;
            if(bluetoothOriginalName.length() > MAX_NAME_SIZE - IDENTIFIER_APP.length()){
                newName = newName.substring(0, MAX_NAME_SIZE - IDENTIFIER_APP.length());
            }
            newName = newName + IDENTIFIER_APP;
            changeBluetoothName(newName);
        }

        IntentFilter actionFoundFilter = new IntentFilter(BluetoothDevice.ACTION_FOUND);
        registerReceiver(actionFoundReceiver, actionFoundFilter);
        IntentFilter discoveryFinishedFilter = new IntentFilter(BluetoothAdapter.ACTION_DISCOVERY_FINISHED);
        registerReceiver(discoveryFinishedReceiver, discoveryFinishedFilter);
        IntentFilter discoveryStartedFilter = new IntentFilter(BluetoothAdapter.ACTION_DISCOVERY_STARTED);
        registerReceiver(discoveryStartedReceiver, discoveryStartedFilter);

        mainChatService.start();

        timer = new Timer();

        showButtonsFragment();
    }

    private void changeBluetoothName(String newName){
        final Handler myTimerHandler = new Handler();
        myTimerHandler.postDelayed(
                new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (bluetoothAdapter.isEnabled())
                        {
                            bluetoothAdapter.setName(newName);
                        }
                        if ((!newName.equalsIgnoreCase(bluetoothAdapter.getName())) && (System.currentTimeMillis() < 10000))
                        {
                            myTimerHandler.postDelayed(this, 500);
                        }
                    }
                }, 500);
    }

    private void showButtonsFragment(){
        if(!isFinishing()){
            FragmentManager fragmentManager = getSupportFragmentManager();
            FragmentTransaction fragmentTransaction = fragmentManager.beginTransaction();
            mainButtonsFragment = new MainButtonsFragment();
            fragmentTransaction.replace(R.id.mainScreen, mainButtonsFragment, "MainButtonsFragment");
            fragmentTransaction.addToBackStack("MainButtonsFragment");
            fragmentTransaction.commit();
        }
    }

    private boolean containsIdetifier(String name){
        return name.contains(IDENTIFIER_APP);
    }

    private void showToastText(String text){
        Toast.makeText(this, text, Toast.LENGTH_LONG).show();
    }

    private void processFinishedSearch(){
        chatDevices = chatDevices.stream().filter(device -> containsIdetifier(device.getNickName())).collect(Collectors.toList());

        if (chatDevices.isEmpty()) {
            new AlertDialog.Builder(this)
                    .setTitle("Nenhum dispositivo encontrado")
                    .setPositiveButton("Desistir", (dialog, id) -> {
                        //DoNothing
                    })
                    .setNegativeButton("Tentar Novamente", (dialog, id) -> findDevices(null))
                    .show();
        }else{
            new AlertDialog.Builder(this)
                    .setTitle("Encontramos um chat em grupo!")
                    .setMessage("Gostaria de se conectar?")
                    .setPositiveButton("Sim", (dialog, id) -> makeConnections())
                    .setNegativeButton("Não", (dialog, id) -> {
                        //DoNothing
                    })
                    .show();
        }
    }

    private void makeConnections(){
        chatDevices.forEach(neighbor -> {
            BluetoothChatService chatService = neighbor.getChatService();
            try {
                if(neighbor.getBluetoothDevice().getBondState() == BluetoothDevice.BOND_NONE){
                    neighbor.getBluetoothDevice().createBond();
                }
            }catch (Exception e){
                Log.e("TAG", Objects.requireNonNull(e.getMessage()));
            }finally {
                chatService.connect(neighbor.getBluetoothDevice(), true);
            }
        });
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, @Nullable Intent data) {
        super.onActivityResult(requestCode, resultCode, data);
        if(resultCode == RESULT_CANCELED){
            showToastText("Tudo bem... ☹️");
        }
        if(resultCode == RESULT_OK){
            showToastText("Obrigado! \uD83D\uDE04");
        }
    }

    @Override
    public void onRequestPermissionsResult(int requestCode, @NonNull String[] permissions, @NonNull int[] grantResults) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults);

        if(requestCode == ENABLE_BLUETOOTH_CODE){
            if(grantResults.length > 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED){
                showToastText("Permissão Concedida :)");
            }else{
                showToastText("Permissão Negada!");
            }
        }
    }

    public void findDevices(View view){
        checkPermissions();

        if(bluetoothAdapter.isDiscovering()){
            bluetoothAdapter.cancelDiscovery();
        }
        bluetoothAdapter.startDiscovery();
    }

    public void disconnect(View view){
       chatDevices.forEach(device-> device.getChatService().stop());
       chatDevices.clear();
       onlineDevices.clear();
    }

    public void openChat(View view){
        showChatFragment();
    }

    public void openInfo(View view){
        AtomicInteger index = new AtomicInteger();
        ArrayList<String> items = new ArrayList<>();
        if(onlineDevices.isEmpty()){
            items.add("O chat está vazio!");
        }else {
            onlineDevices.forEach(device -> {
                    index.getAndIncrement();
                    items.add(index + " - " + removeIdentifier(device));
                }
            );
        }

        String[] itemsConverted = items.toArray(new String[0]);
        new AlertDialog.Builder(mainActivityContext)
            .setTitle("Pessoas Online:")
            .setPositiveButton ("Fechar", (dialog, id) -> dialog.dismiss())
            .setItems(itemsConverted, (dialog, id) -> {
                //DoNothing
            })
            .show();
    }

    private String getFirstNameWithoutIdentifier(){
        String myName = bluetoothAdapter.getName();
        String[] arr = myName.split(" ", 2);
        String firstWordOfName = arr[0];
        return removeIdentifier(firstWordOfName);
    }

    private String getFullNameWithoutIdentifier(){
        String myName = bluetoothAdapter.getName();
        return removeIdentifier(myName);
    }

    private void saveUuidByMessageAndCreateTaskAutoRemove(String message){
        String uuid = converter.getUuidFromMessage(message);
        recentUuids.add(uuid);
        timer.schedule(createTimerTaskRemoveUuidFromList(), MESSAGE_UUID_LIFETIME_MS);
    }

    private void saveUuidInRecentsAndCreateTaskAutoRemove(String uuid){
        recentUuids.add(uuid);
        timer.schedule(createTimerTaskRemoveUuidFromList(), MESSAGE_UUID_LIFETIME_MS);
    }

    private String generateNewUuid(){
        return UUID.randomUUID().toString();
    }

    private TimerTask createTimerTaskRemoveUuidFromList(){
        return new TimerTask() {
            @Override
            public void run() {
                recentUuids.remove(0);
            }
        };
    }

    private String removeIdentifier(String text) {
        return text.contains(IDENTIFIER_APP) ? text.substring(0, text.length() - 3) : text;
    }

    public void becomeVisible(View view){
        Intent discoverableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_DISCOVERABLE);
        discoverableIntent.putExtra(BluetoothAdapter.EXTRA_DISCOVERABLE_DURATION, 300);
        startActivity(discoverableIntent);
    }

    private void showChatFragment(){
        if(!isFinishing()){
            FragmentManager fragmentManager = getSupportFragmentManager();
            FragmentTransaction fragmentTransaction = fragmentManager.beginTransaction();
            chatFragment.setCommunicationListener(this);
            fragmentTransaction.replace(R.id.mainScreen, chatFragment, "ChatFragment");
            fragmentTransaction.addToBackStack("ChatFragment");
            fragmentTransaction.commit();
        }
    }

    private final BroadcastReceiver actionFoundReceiver = new BroadcastReceiver() {
        public void onReceive(Context context, Intent intent) {
            String action = intent.getAction();

            if (BluetoothDevice.ACTION_FOUND.equals(action)) {
                BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                if(Objects.nonNull(device) && Objects.nonNull(device.getName())){
                    String deviceName = device.getName();
                    if(!containsIdetifier(deviceName)){
                        return;
                    }

                    boolean alreadyAdded = chatDevices.stream().anyMatch(
                            deviceInList -> deviceInList.getBluetoothDevice().getAddress().equals(device.getAddress())
                    );
                    if(!alreadyAdded) {
                        chatDevices.add(
                                Device.builder()
                                        .nickName(device.getName())
                                        .bluetoothDevice(device)
                                        .chatService(new BluetoothChatService(mainActivityContext, blueChatHandler))
                                        .build()
                        );
                    }
                }
            }
        }
    };

    private final BroadcastReceiver discoveryFinishedReceiver = new BroadcastReceiver() {
        public void onReceive(Context context, Intent intent) {
            String action = intent.getAction();
            if (BluetoothAdapter.ACTION_DISCOVERY_FINISHED.equals(action)) {
                showToastText("TERMINADO!");
                processFinishedSearch();
            }
        }
    };

    private final BroadcastReceiver discoveryStartedReceiver = new BroadcastReceiver() {
        public void onReceive(Context context, Intent intent) {
            String action = intent.getAction();
            if (BluetoothAdapter.ACTION_DISCOVERY_STARTED.equals(action)) {
                showToastText("Buscando...");
            }
        }
    };

    @Override
    public void onCommunication(String message) {
        if(chatDevices.isEmpty()){
            chatFragment.showSnackEmptyChat();
            return;
        }

        String myName = bluetoothAdapter.getName();
        String[] arr = myName.split(" ", 2);
        String firstWordOfName = arr[0];
        String nameWithoutIdentifier = removeIdentifier(firstWordOfName);
        String uuid = generateNewUuid();

        String mountedMessage = converter.mountCompleteMessage(
                MessageTypes.CHAT_ORIGIN, uuid, nameWithoutIdentifier, message
        );

        distributeMessage(mountedMessage);
    }


    @Override
    protected void onStop() {
        super.onStop();
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        unregisterReceiver(discoveryStartedReceiver);
        unregisterReceiver(actionFoundReceiver);
        unregisterReceiver(discoveryFinishedReceiver);
    }

    private void checkBluetooth(){
        checkBluetoothSupport();
        checkBluetoothEnabled();
    }

    private void checkBluetoothSupport(){
        if (bluetoothAdapter == null) {
            Log.i("INFO","Sem suporte a Bluetooth!");
            showToastText("Bluetooth não suportado!");
            System.exit(1);
        }else{
            Log.i("INFO","Suporte a Bluetooth detectado!");
        }
    }

    private void checkBluetoothEnabled(){
        if (!bluetoothAdapter.isEnabled()) {
            Intent enableBtIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            startActivityForResult(enableBtIntent, ENABLE_BLUETOOTH_CODE);
        }
    }

    private void checkPermissions(){
        if (ContextCompat.checkSelfPermission(mainActivityContext, Manifest.permission.BLUETOOTH)!= PackageManager.PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(MainActivity.this, new String[]{Manifest.permission.BLUETOOTH}, BLUETOOTH_RQ);
        }
        if (ContextCompat.checkSelfPermission(mainActivityContext, Manifest.permission.BLUETOOTH_ADMIN)!= PackageManager.PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(MainActivity.this, new String[]{Manifest.permission.BLUETOOTH_ADMIN}, BLUETOOTH_ADMIN_RQ);
        }
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q) {
            if (ContextCompat.checkSelfPermission(mainActivityContext, Manifest.permission.ACCESS_FINE_LOCATION)!= PackageManager.PERMISSION_GRANTED) {
                ActivityCompat.requestPermissions(MainActivity.this, new String[]{Manifest.permission.ACCESS_FINE_LOCATION}, FINE_LOCATION_RQ);
            }
        }else{
            if (ContextCompat.checkSelfPermission(mainActivityContext, Manifest.permission.ACCESS_COARSE_LOCATION)!= PackageManager.PERMISSION_GRANTED) {
                ActivityCompat.requestPermissions(MainActivity.this, new String[]{Manifest.permission.ACCESS_COARSE_LOCATION}, COARSE_LOCATION_RQ);
            }
        }
    }
}